"""Timestamp writer for keeping track of bulk optimizations."""
from treestamps.tree.dump import DumpMixin
from treestamps.tree.load import LoadMixin
from treestamps.tree.set import SetMixin


class Treestamps(SetMixin, LoadMixin, DumpMixin):
    """Treestamps object to hold settings and caches."""



"""Treestamps Config methods."""
from collections.abc import Iterable
from dataclasses import dataclass
from typing import Optional


def normalize_config(config: Optional[dict]) -> Optional[dict]:
    """Recursively convert iterables into sorted unique lists."""
    if config is None:
        return None
    new_config: dict = {}
    for key, value in config.items():
        if isinstance(value, (list, tuple, set, frozenset)):
            new_config[key] = sorted(frozenset(value))
        elif isinstance(value, dict):
            new_config[key] = normalize_config(value)
        else:
            new_config[key] = value

    return new_config


DEFAULT_CONFIG = normalize_config({"ignore": frozenset(), "symlinks": True})


@dataclass
class CommonConfig:
    """Common Config meant to be subclassed."""

    program_name: str
    verbose: int = 0
    symlinks: bool = True
    ignore: Iterable[str] = frozenset()
    check_config: bool = True
    program_config: Optional[dict] = None
    program_config_keys: Iterable[str] = frozenset()

    def __post_init__(self):
        """Fix types and normalize program config dict."""
        self.ignore = frozenset(self.ignore)
        self.program_config_keys = frozenset(self.program_config_keys)

        # Filter dict by keys
        if self.program_config:

            def filter_func(pair):
                return pair[0] in self.program_config_keys

            self.program_config = dict(filter(filter_func, self.program_config.items()))

        self.program_config = normalize_config(self.program_config)



"""Load methods."""
from pathlib import Path

from termcolor import cprint

from treestamps.config import DEFAULT_CONFIG, normalize_config
from treestamps.tree.get import GetMixin


class LoadMixin(GetMixin):
    """Load methods."""

    def _is_path_skipped(self, path: Path) -> bool:
        """Return if path is ignored or not allowed because symlink."""
        return any(path.match(ignore_glob) for ignore_glob in self._config.ignore) or (
            not self._config.symlinks and path.is_symlink()
        )

    def _load_timestamps_file_config_matches(self, yaml):
        """Return if the configured and loaded configs match."""
        yaml_ts_config = yaml.pop(self._TREESTAMPS_CONFIG_TAG, DEFAULT_CONFIG)
        yaml_program_config = yaml.pop(self._CONFIG_TAG, None)
        if not self._config.check_config:
            return True

        ts_config = self._get_treestamps_config_dict()
        if yaml_ts_config != ts_config:
            return False

        yaml_program_config = normalize_config(yaml_program_config)
        if self._config.program_config != yaml_program_config:
            # Only load timestamps for comparable configs
            return False

        return True

    def _get_absolute_entry_path(self, root: Path, path_str) -> Path:
        """Get the path from a timestamp file relative to this treestamp's root_dir."""
        abs_root = self._get_absolute_path(self.root_dir, root)
        if not abs_root:
            raise ValueError
        abs_path = self._get_absolute_path(abs_root, path_str)
        if not abs_path:
            raise ValueError
        return abs_path

    def _load_timestamp_entry(self, root: Path, path_str: str, ts: float) -> None:
        """Load a single timestamp entry into the cache."""
        try:
            abs_path = self._get_absolute_entry_path(root, path_str)
            old_ts = self.get(abs_path)
            if abs_path not in self._timestamps or old_ts is None or ts > old_ts:
                self._timestamps[abs_path] = ts
        except Exception as exc:
            cprint(f"WARNING: Invalid timestamp for {path_str}: {ts} {exc}", "yellow")

    def _load_timestamps_file(self, timestamps_path: Path) -> None:
        """Load timestamps from a file."""
        if not timestamps_path.is_file():
            return

        try:
            yaml = self._YAML.load(timestamps_path)
            if not yaml:
                return

            # pops off config entries.
            if not self._load_timestamps_file_config_matches(yaml):
                return

            # WAL
            wal = yaml.pop(self._WAL_TAG, [])

            # What's left are timestamps
            entries = list(yaml.items())

            # Wal entries afterwards
            for entry in wal:
                try:
                    for path_str, ts in entry.items():
                        entries += [(path_str, ts)]
                except Exception as exc:
                    cprint(f"WAL entry read: {exc}", "yellow")

            for path_str, ts in entries:
                self._load_timestamp_entry(timestamps_path.parent, path_str, ts)
        except Exception as exc:
            cprint(f"ERROR: parsing timestamps file: {timestamps_path} {exc}", "red")

    def _consume_child_timestamps(self, path: Path) -> None:
        """Consume a child timestamp and add its values to our root."""
        try:
            if not path.is_file() or self._is_path_skipped(path):
                return
            self._load_timestamps_file(path)
            if path != self._dump_path:
                self._consumed_paths.add(path)
            if self._config.verbose:
                cprint(f"Read timestamps from {path}")
        except Exception as exc:
            cprint(f"WARNING: reading child timestamps {exc}", "yellow")

    def _consume_all_child_timestamps(
        self, path: Path, do_consume_children=True
    ) -> None:
        """Recursively consume all timestamps and wal files."""
        try:
            if not path.is_dir() or self._is_path_skipped(path):
                return
            for name in (self._filename, self._wal_filename):
                self._consume_child_timestamps(path / name)
            if do_consume_children:
                for dir_entry in path.iterdir():
                    self._consume_all_child_timestamps(dir_entry)
        except Exception as exc:
            cprint(f"WARNING: reading all child timstamps {exc}", "yellow")

    def _load_parent_timestamps(self, path: Path) -> None:
        """Recursively load timestamps from all parents."""
        if path.parent == path.parent.parent or self._is_path_skipped(path):
            return
        parent = path.parent
        self._load_timestamps_file(parent / self._filename)
        self._load_timestamps_file(parent / self._wal_filename)
        self._load_parent_timestamps(parent)

    def load(self) -> None:
        """Load all timestamps."""
        self._load_parent_timestamps(self.root_dir)
        do_consume_children = self._config.path.is_dir()
        self._consume_all_child_timestamps(self.root_dir, do_consume_children)



"""A dict of Treestamps."""
from collections.abc import Iterable
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import Union

from termcolor import cprint

from treestamps.config import CommonConfig
from treestamps.tree import Treestamps
from treestamps.tree.common import TreestampsConfig


@dataclass
class GrovestampsConfig(CommonConfig):
    """Grovestamps config."""

    paths: Iterable[Union[str, Path]] = ()


class Grovestamps(dict):
    """A path keyed dict of Treestamps."""

    def _order_paths(self) -> tuple[Path]:
        """Return ordered deduplicated list of paths.

        This order creates dir based treestamps before files so dirs get children
        recursed and files only don't.
        """
        dirs = set()
        files = set()
        for path_str in self._config.paths:
            path = Path(path_str)
            if not self._config.symlinks and path.is_symlink():
                continue
            if path.is_dir():
                dirs.add(path)
            else:
                files.add(path)
        return tuple(sorted(dirs) + sorted(files))

    def __init__(self, config: GrovestampsConfig) -> None:
        """Create a dictionary of Treestamps keyed with paths."""
        self._config = config

        ordered_paths = self._order_paths()

        config_dict = asdict(self._config)
        config_dict.pop("paths", None)

        for top_path in ordered_paths:
            root_dir = Treestamps.get_dir(top_path)
            if root_dir in self:
                continue
            tree_config = TreestampsConfig(**config_dict, path=top_path)
            ts = Treestamps(tree_config)
            self[root_dir] = ts
            ts.load()

    def dump(self) -> None:
        """Dump all treestamps."""
        for top_path, treestamps in self.items():
            if self._config.verbose:
                cprint(f"Saving timestamps for {top_path}")
            treestamps.dump()
